@@ -1,18 +1,14 @@
-// This is an open source non-commercial project. Dear PVS-Studio, please check
-// it. PVS-Studio Static Code Analyzer for C, C++ and C#: http://www.viva64.com
-
 #include <assert.h>
-#include <stdint.h>
 #include <stdbool.h>
-#include <stdlib.h>
-
+#include <stddef.h>
 #include <uv.h>
 
-#include "nvim/log.h"
-#include "nvim/event/loop.h"
+#include "nvim/event/defs.h"
+#include "nvim/event/stream.h"
 #include "nvim/event/wstream.h"
-#include "nvim/vim.h"
+#include "nvim/macros_defs.h"
 #include "nvim/memory.h"
+#include "nvim/types_defs.h"
 
 #define DEFAULT_MAXMEM 1024 * 1024 * 2000
 
@@ -107,7 +103,7 @@ err:
 
 /// Creates a WBuffer object for holding output data. Instances of this
 /// object can be reused across Stream instances, and the memory is freed
-/// automatically when no longer needed(it tracks the number of references
+/// automatically when no longer needed (it tracks the number of references
 /// internally)
 ///
 /// @param data Data stored by the WBuffer
@@ -115,12 +111,9 @@ err:
 /// @param refcount The number of references for the WBuffer. This will be used
 ///        by Stream instances to decide when a WBuffer should be freed.
 /// @param cb Pointer to function that will be responsible for freeing
-///        the buffer data(passing 'free' will work as expected).
+///        the buffer data (passing `xfree` will work as expected).
 /// @return The allocated WBuffer instance
-WBuffer *wstream_new_buffer(char *data,
-                            size_t size,
-                            size_t refcount,
-                            wbuffer_data_finalizer cb)
+WBuffer *wstream_new_buffer(char *data, size_t size, size_t refcount, wbuffer_data_finalizer cb)
   FUNC_ATTR_NONNULL_ARG(1)
 {
   WBuffer *rv = xmalloc(sizeof(WBuffer));
