@@ -1,43 +1,46 @@
-local helpers = require('test.functional.helpers')(after_each)
-local lsp_helpers = require('test.functional.plugin.lsp.helpers')
+local t = require('test.testutil')
+local n = require('test.functional.testnvim')()
 local Screen = require('test.functional.ui.screen')
+local t_lsp = require('test.functional.plugin.lsp.testutil')
 
-local command = helpers.command
-local dedent = helpers.dedent
-local eq = helpers.eq
-local exec_lua = helpers.exec_lua
-local feed = helpers.feed
-local feed_command = helpers.feed_command
-local insert = helpers.insert
-local matches = helpers.matches
+local command = n.command
+local dedent = t.dedent
+local eq = t.eq
+local exec_lua = n.exec_lua
+local feed = n.feed
+local feed_command = n.feed_command
+local insert = n.insert
+local matches = t.matches
+local api = n.api
 
-local clear_notrace = lsp_helpers.clear_notrace
-local create_server_definition = lsp_helpers.create_server_definition
+local clear_notrace = t_lsp.clear_notrace
+local create_server_definition = t_lsp.create_server_definition
 
 before_each(function()
   clear_notrace()
 end)
 
 after_each(function()
-  exec_lua("vim.api.nvim_exec_autocmds('VimLeavePre', { modeline = false })")
+  api.nvim_exec_autocmds('VimLeavePre', { modeline = false })
 end)
 
 describe('semantic token highlighting', function()
-
   local screen
   before_each(function()
     screen = Screen.new(40, 16)
     screen:attach()
     screen:set_default_attr_ids {
-      [1] = { bold = true, foreground = Screen.colors.Blue1 };
-      [2] = { foreground = Screen.colors.DarkCyan };
-      [3] = { foreground = Screen.colors.SlateBlue };
-      [4] = { bold = true, foreground = Screen.colors.SeaGreen };
-      [5] = { foreground = tonumber('0x6a0dad') };
-      [6] = { foreground = Screen.colors.Blue1 };
-      [7] = { bold = true, foreground = Screen.colors.DarkCyan };
-      [8] = { bold = true, foreground = Screen.colors.SlateBlue };
-      [9] = { bold = true, foreground = tonumber('0x6a0dad') };
+      [1] = { bold = true, foreground = Screen.colors.Blue1 },
+      [2] = { foreground = Screen.colors.DarkCyan },
+      [3] = { foreground = Screen.colors.SlateBlue },
+      [4] = { bold = true, foreground = Screen.colors.SeaGreen },
+      [5] = { foreground = tonumber('0x6a0dad') },
+      [6] = { foreground = Screen.colors.Blue1 },
+      [7] = { bold = true, foreground = Screen.colors.DarkCyan },
+      [8] = { bold = true, foreground = Screen.colors.SlateBlue },
+      [9] = { bold = true, foreground = tonumber('0x6a0dad') },
+      [10] = { bold = true, foreground = Screen.colors.Brown },
+      [11] = { foreground = Screen.colors.Magenta1 },
     }
     command([[ hi link @lsp.type.namespace Type ]])
     command([[ hi link @lsp.type.function Special ]])
@@ -81,7 +84,8 @@ describe('semantic token highlighting', function()
 
     before_each(function()
       exec_lua(create_server_definition)
-      exec_lua([[
+      exec_lua(
+        [[
         local legend, response, edit_response = ...
         server = _create_server({
           capabilities = {
@@ -99,19 +103,25 @@ describe('semantic token highlighting', function()
             end,
           }
         })
-      ]], legend, response, edit_response)
+      ]],
+        legend,
+        response,
+        edit_response
+      )
     end)
 
     it('buffer is highlighted when attached', function()
       exec_lua([[
         bufnr = vim.api.nvim_get_current_buf()
         vim.api.nvim_win_set_buf(0, bufnr)
+        vim.bo[bufnr].filetype = 'some-filetype'
         client_id = vim.lsp.start({ name = 'dummy', cmd = server.cmd })
       ]])
 
       insert(text)
 
-      screen:expect { grid = [[
+      screen:expect {
+        grid = [[
         #include <iostream>                     |
                                                 |
         int {8:main}()                              |
@@ -124,11 +134,10 @@ describe('semantic token highlighting', function()
         {6:#endif}                                  |
         }                                       |
         ^}                                       |
-        {1:~                                       }|
-        {1:~                                       }|
-        {1:~                                       }|
+        {1:~                                       }|*3
                                                 |
-      ]] }
+      ]],
+      }
     end)
 
     it('use LspTokenUpdate and highlight_token', function()
@@ -150,7 +159,8 @@ describe('semantic token highlighting', function()
 
       insert(text)
 
-      screen:expect { grid = [[
+      screen:expect {
+        grid = [[
         #include <iostream>                     |
                                                 |
         int {9:main}()                              |
@@ -163,12 +173,10 @@ describe('semantic token highlighting', function()
         {6:#endif}                                  |
         }                                       |
         ^}                                       |
-        {1:~                                       }|
-        {1:~                                       }|
-        {1:~                                       }|
+        {1:~                                       }|*3
                                                 |
-      ]] }
-
+      ]],
+      }
     end)
 
     it('buffer is unhighlighted when client is detached', function()
@@ -185,7 +193,8 @@ describe('semantic token highlighting', function()
         vim.lsp.buf_detach_client(bufnr, client_id)
       ]])
 
-      screen:expect { grid = [[
+      screen:expect {
+        grid = [[
         #include <iostream>                     |
                                                 |
         int main()                              |
@@ -198,29 +207,30 @@ describe('semantic token highlighting', function()
         #endif                                  |
         }                                       |
         ^}                                       |
-        {1:~                                       }|
-        {1:~                                       }|
-        {1:~                                       }|
+        {1:~                                       }|*3
                                                 |
-      ]] }
+      ]],
+      }
     end)
 
-    it('buffer is highlighted and unhighlighted when semantic token highlighting is started and stopped'
-      , function()
-      exec_lua([[
+    it(
+      'buffer is highlighted and unhighlighted when semantic token highlighting is started and stopped',
+      function()
+        exec_lua([[
         bufnr = vim.api.nvim_get_current_buf()
         vim.api.nvim_win_set_buf(0, bufnr)
         client_id = vim.lsp.start({ name = 'dummy', cmd = server.cmd })
       ]])
 
-      insert(text)
+        insert(text)
 
-      exec_lua([[
+        exec_lua([[
         vim.notify = function() end
         vim.lsp.semantic_tokens.stop(bufnr, client_id)
       ]])
 
-      screen:expect { grid = [[
+        screen:expect {
+          grid = [[
         #include <iostream>                     |
                                                 |
         int main()                              |
@@ -233,17 +243,17 @@ describe('semantic token highlighting', function()
         #endif                                  |
         }                                       |
         ^}                                       |
-        {1:~                                       }|
-        {1:~                                       }|
-        {1:~                                       }|
+        {1:~                                       }|*3
                                                 |
-      ]] }
+      ]],
+        }
 
-      exec_lua([[
+        exec_lua([[
         vim.lsp.semantic_tokens.start(bufnr, client_id)
       ]])
 
-      screen:expect { grid = [[
+        screen:expect {
+          grid = [[
         #include <iostream>                     |
                                                 |
         int {8:main}()                              |
@@ -256,12 +266,12 @@ describe('semantic token highlighting', function()
         {6:#endif}                                  |
         }                                       |
         ^}                                       |
-        {1:~                                       }|
-        {1:~                                       }|
-        {1:~                                       }|
+        {1:~                                       }|*3
                                                 |
-      ]] }
-    end)
+      ]],
+        }
+      end
+    )
 
     it('buffer is re-highlighted when force refreshed', function()
       exec_lua([[
@@ -272,7 +282,8 @@ describe('semantic token highlighting', function()
 
       insert(text)
 
-      screen:expect { grid = [[
+      screen:expect {
+        grid = [[
         #include <iostream>                     |
                                                 |
         int {8:main}()                              |
@@ -285,17 +296,17 @@ describe('semantic token highlighting', function()
         {6:#endif}                                  |
         }                                       |
         ^}                                       |
-        {1:~                                       }|
-        {1:~                                       }|
-        {1:~                                       }|
+        {1:~                                       }|*3
                                                 |
-      ]] }
+      ]],
+      }
 
       exec_lua([[
         vim.lsp.semantic_tokens.force_refresh(bufnr)
       ]])
 
-      screen:expect { grid = [[
+      screen:expect {
+        grid = [[
         #include <iostream>                     |
                                                 |
         int {8:main}()                              |
@@ -308,11 +319,11 @@ describe('semantic token highlighting', function()
         {6:#endif}                                  |
         }                                       |
         ^}                                       |
-        {1:~                                       }|
-        {1:~                                       }|
-        {1:~                                       }|
+        {1:~                                       }|*3
                                                 |
-      ]], unchanged = true }
+      ]],
+        unchanged = true,
+      }
 
       local messages = exec_lua('return server.messages')
       local token_request_count = 0
@@ -351,7 +362,8 @@ describe('semantic token highlighting', function()
       ]])
 
       insert(text)
-      screen:expect { grid = [[
+      screen:expect {
+        grid = [[
         #include <iostream>                     |
                                                 |
         int {8:main}()                              |
@@ -364,14 +376,14 @@ describe('semantic token highlighting', function()
         {6:#endif}                                  |
         }                                       |
         ^}                                       |
-        {1:~                                       }|
-        {1:~                                       }|
-        {1:~                                       }|
+        {1:~                                       }|*3
                                                 |
-      ]] }
+      ]],
+      }
       feed_command('%s/int x/int x()/')
       feed_command('noh')
-      screen:expect { grid = [[
+      screen:expect {
+        grid = [[
         #include <iostream>                     |
                                                 |
         int {8:main}()                              |
@@ -382,13 +394,11 @@ describe('semantic token highlighting', function()
         {6:#else}                                   |
         {6:    printf("%d\n", x);}                  |
         {6:#endif}                                  |
-        }                                       |
-        }                                       |
-        {1:~                                       }|
-        {1:~                                       }|
-        {1:~                                       }|
+        }                                       |*2
+        {1:~                                       }|*3
         :noh                                    |
-      ]] }
+      ]],
+      }
     end)
 
     it('prevents starting semantic token highlighting with invalid conditions', function()
@@ -399,7 +409,7 @@ describe('semantic token highlighting', function()
         notifications = {}
         vim.notify = function(...) table.insert(notifications, 1, {...}) end
       ]])
-      eq(false, exec_lua("return vim.lsp.buf_is_attached(bufnr, client_id)"))
+      eq(false, exec_lua('return vim.lsp.buf_is_attached(bufnr, client_id)'))
 
       insert(text)
 
@@ -416,7 +426,8 @@ describe('semantic token highlighting', function()
       matches('%[LSP%] No client with id %d', notifications[1][1])
     end)
 
-    it('opt-out: does not activate semantic token highlighting if disabled in client attach',
+    it(
+      'opt-out: does not activate semantic token highlighting if disabled in client attach',
       function()
         exec_lua([[
           bufnr = vim.api.nvim_get_current_buf()
@@ -429,11 +440,12 @@ describe('semantic token highlighting', function()
             end),
           })
         ]])
-        eq(true, exec_lua("return vim.lsp.buf_is_attached(bufnr, client_id)"))
+        eq(true, exec_lua('return vim.lsp.buf_is_attached(bufnr, client_id)'))
 
         insert(text)
 
-        screen:expect { grid = [[
+        screen:expect {
+          grid = [[
           #include <iostream>                     |
                                                   |
           int main()                              |
@@ -446,11 +458,10 @@ describe('semantic token highlighting', function()
           #endif                                  |
           }                                       |
           ^}                                       |
-          {1:~                                       }|
-          {1:~                                       }|
-          {1:~                                       }|
+          {1:~                                       }|*3
                                                   |
-        ]] }
+        ]],
+        }
 
         local notifications = exec_lua([[
           local notifications = {}
@@ -460,7 +471,8 @@ describe('semantic token highlighting', function()
         ]])
         eq('[LSP] Server does not support semantic tokens', notifications[1][1])
 
-        screen:expect { grid = [[
+        screen:expect {
+          grid = [[
           #include <iostream>                     |
                                                   |
           int main()                              |
@@ -473,14 +485,15 @@ describe('semantic token highlighting', function()
           #endif                                  |
           }                                       |
           ^}                                       |
-          {1:~                                       }|
-          {1:~                                       }|
-          {1:~                                       }|
+          {1:~                                       }|*3
                                                   |
-          ]], unchanged = true }
-      end)
+          ]],
+          unchanged = true,
+        }
+      end
+    )
 
-  it('ignores null responses from the server', function()
+    it('ignores null responses from the server', function()
       exec_lua([[
         local legend, response, edit_response = ...
         server2 = _create_server({
@@ -502,11 +515,12 @@ describe('semantic token highlighting', function()
         vim.api.nvim_win_set_buf(0, bufnr)
         client_id = vim.lsp.start({ name = 'dummy', cmd = server2.cmd })
       ]])
-      eq(true, exec_lua("return vim.lsp.buf_is_attached(bufnr, client_id)"))
+      eq(true, exec_lua('return vim.lsp.buf_is_attached(bufnr, client_id)'))
 
       insert(text)
 
-      screen:expect { grid = [[
+      screen:expect {
+        grid = [[
         #include <iostream>                     |
                                                 |
         int main()                              |
@@ -519,15 +533,15 @@ describe('semantic token highlighting', function()
         #endif                                  |
         }                                       |
         ^}                                       |
-        {1:~                                       }|
-        {1:~                                       }|
-        {1:~                                       }|
+        {1:~                                       }|*3
                                                 |
-      ]] }
+      ]],
+      }
     end)
 
     it('does not send delta requests if not supported by server', function()
-      exec_lua([[
+      exec_lua(
+        [[
         local legend, response, edit_response = ...
         server2 = _create_server({
           capabilities = {
@@ -548,10 +562,15 @@ describe('semantic token highlighting', function()
         bufnr = vim.api.nvim_get_current_buf()
         vim.api.nvim_win_set_buf(0, bufnr)
         client_id = vim.lsp.start({ name = 'dummy', cmd = server2.cmd })
-      ]], legend, response, edit_response)
+      ]],
+        legend,
+        response,
+        edit_response
+      )
 
       insert(text)
-      screen:expect { grid = [[
+      screen:expect {
+        grid = [[
         #include <iostream>                     |
                                                 |
         int {8:main}()                              |
@@ -564,18 +583,18 @@ describe('semantic token highlighting', function()
         {6:#endif}                                  |
         }                                       |
         ^}                                       |
-        {1:~                                       }|
-        {1:~                                       }|
-        {1:~                                       }|
+        {1:~                                       }|*3
                                                 |
-      ]] }
+      ]],
+      }
       feed_command('%s/int x/int x()/')
       feed_command('noh')
 
       -- the highlights don't change because our fake server sent the exact
       -- same result for the same method (the full request). "x" would have
       -- changed to highlight index 3 had we sent a delta request
-      screen:expect { grid = [[
+      screen:expect {
+        grid = [[
         #include <iostream>                     |
                                                 |
         int {8:main}()                              |
@@ -586,13 +605,11 @@ describe('semantic token highlighting', function()
         {6:#else}                                   |
         {6:    printf("%d\n", x);}                  |
         {6:#endif}                                  |
-        }                                       |
-        }                                       |
-        {1:~                                       }|
-        {1:~                                       }|
-        {1:~                                       }|
+        }                                       |*2
+        {1:~                                       }|*3
         :noh                                    |
-      ]] }
+      ]],
+      }
       local messages = exec_lua('return server2.messages')
       local token_request_count = 0
       for _, message in ipairs(messages) do
@@ -630,24 +647,13 @@ describe('semantic token highlighting', function()
           },
         },
         expected_screen = function()
-          screen:expect{grid=[[
+          screen:expect {
+            grid = [[
             char* {7:foo} = "\n"^;                       |
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
+            {1:~                                       }|*14
                                                     |
-          ]]}
+          ]],
+          }
         end,
       },
       {
@@ -762,7 +768,8 @@ int main()
           },
         },
         expected_screen = function()
-          screen:expect{grid=[[
+          screen:expect {
+            grid = [[
             #include <iostream>                     |
             int {8:main}()                              |
             {                                       |
@@ -773,13 +780,10 @@ int main()
             {6:    comment}                             |
             {6:  #endif}                                |
             ^}                                       |
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
+            {1:~                                       }|*5
                                                     |
-          ]]}
+          ]],
+          }
         end,
       },
       {
@@ -823,34 +827,27 @@ b = "as"]],
           },
         },
         expected_screen = function()
-          screen:expect{grid=[[
+          screen:expect {
+            grid = [[
             {6:-- comment}                              |
             local {7:a} = 1                             |
             {2:b} = "as^"                                |
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
+            {1:~                                       }|*12
                                                     |
-          ]]}
+          ]],
+          }
         end,
       },
       {
         it = 'rust-analyzer',
         text = [[pub fn main() {
-  break rust;
-  /// what?
+    println!("Hello world!");
+    break rust;
+    /// what?
 }
 ]],
-        response = [[{"data": [0, 0, 3, 1, 0, 0, 4, 2, 1, 0, 0, 3, 4, 14, 524290, 0, 4, 1, 45, 0, 0, 1, 1, 45, 0, 0, 2, 1, 26, 0, 1, 4, 5, 1, 8192, 0, 6, 4, 52, 0, 0, 4, 1, 48, 0, 1, 4, 9, 0, 1, 1, 0, 1, 26, 0], "resultId": "1"}]],
+        response = [[{"data": [0, 0, 3, 1, 0, 0, 4, 2, 1, 0, 0, 3, 4, 14, 524290, 0, 4, 1, 45, 0, 0, 1, 1, 45, 0, 0, 2, 1, 26, 0, 1, 4, 8, 17, 0, 0, 8, 1, 45, 0, 0, 1, 14, 2, 0, 0, 14, 1, 45, 0, 0, 1, 1, 48, 0, 1, 4, 5, 1, 8192, 0, 6, 4, 52, 0, 0, 4, 1, 48, 0, 1, 4, 9, 0, 1, 1, 0, 1, 26, 0 ], "resultId": "1"}]],
+
         legend = [[{
         "tokenTypes": [
           "comment", "keyword", "string", "number", "regexp", "operator", "namespace", "type", "struct", "class", "interface", "enum", "enumMember", "typeParameter", "function", "method", "property", "macro", "variable",
@@ -913,6 +910,46 @@ b = "as"]],
           },
           {
             line = 1,
+            modifiers = {},
+            start_col = 4,
+            end_col = 12,
+            type = 'macro', -- println!
+            marked = true,
+          },
+          {
+            line = 1,
+            modifiers = {},
+            start_col = 12,
+            end_col = 13,
+            type = 'parenthesis',
+            marked = true,
+          },
+          {
+            line = 1,
+            modifiers = {},
+            start_col = 13,
+            end_col = 27,
+            type = 'string', -- "Hello world!"
+            marked = true,
+          },
+          {
+            line = 1,
+            modifiers = {},
+            start_col = 27,
+            end_col = 28,
+            type = 'parenthesis',
+            marked = true,
+          },
+          {
+            line = 1,
+            modifiers = {},
+            start_col = 28,
+            end_col = 29,
+            type = 'semicolon',
+            marked = true,
+          },
+          {
+            line = 2,
             modifiers = { controlFlow = true },
             start_col = 4,
             end_col = 9, -- break
@@ -920,31 +957,31 @@ b = "as"]],
             marked = true,
           },
           {
-            line = 1,
+            line = 2,
             modifiers = {},
             start_col = 10,
-            end_col = 13, -- rust
+            end_col = 14, -- rust
             type = 'unresolvedReference',
             marked = true,
           },
           {
-            line = 1,
+            line = 2,
             modifiers = {},
-            start_col = 13,
-            end_col = 13,
+            start_col = 14,
+            end_col = 15,
             type = 'semicolon',
             marked = true,
           },
           {
-            line = 2,
+            line = 3,
             modifiers = { documentation = true },
             start_col = 4,
-            end_col = 11,
+            end_col = 13,
             type = 'comment', -- /// what?
             marked = true,
           },
           {
-            line = 3,
+            line = 4,
             modifiers = {},
             start_col = 0,
             end_col = 1,
@@ -953,30 +990,25 @@ b = "as"]],
           },
         },
         expected_screen = function()
-          screen:expect{grid=[[
-            pub fn {8:main}() {                         |
-              break rust;                           |
-              //{6:/ what?}                             |
+          screen:expect {
+            grid = [[
+            {10:pub} {10:fn} {8:main}() {                         |
+                {5:println!}({11:"Hello world!"});           |
+                {10:break} rust;                         |
+                {6:/// what?}                           |
             }                                       |
             ^                                        |
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
+            {1:~                                       }|*9
                                                     |
-          ]]}
+          ]],
+          }
         end,
       },
     }) do
       it(test.it, function()
         exec_lua(create_server_definition)
-        exec_lua([[
+        exec_lua(
+          [[
           local legend, resp = ...
           server = _create_server({
             capabilities = {
@@ -994,7 +1026,10 @@ b = "as"]],
           bufnr = vim.api.nvim_get_current_buf()
           vim.api.nvim_win_set_buf(0, bufnr)
           client_id = vim.lsp.start({ name = 'dummy', cmd = server.cmd })
-        ]], test.legend, test.response)
+        ]],
+          test.legend,
+          test.response
+        )
 
         insert(test.text)
 
@@ -1036,7 +1071,7 @@ b = "as"]],
             end_col = 9,
             type = 'variable',
             marked = true,
-          }
+          },
         },
         expected2 = {
           {
@@ -1049,47 +1084,26 @@ b = "as"]],
             end_col = 9,
             type = 'variable',
             marked = true,
-          }
+          },
         },
         expected_screen1 = function()
-        screen:expect{grid=[[
+          screen:expect {
+            grid = [[
           char* {7:foo} = "\n"^;                       |
-          {1:~                                       }|
-          {1:~                                       }|
-          {1:~                                       }|
-          {1:~                                       }|
-          {1:~                                       }|
-          {1:~                                       }|
-          {1:~                                       }|
-          {1:~                                       }|
-          {1:~                                       }|
-          {1:~                                       }|
-          {1:~                                       }|
-          {1:~                                       }|
-          {1:~                                       }|
-          {1:~                                       }|
+          {1:~                                       }|*14
                                                   |
-        ]]}
+        ]],
+          }
         end,
         expected_screen2 = function()
-          screen:expect{grid=[[
+          screen:expect {
+            grid = [[
             ^                                        |
             char* {7:foo} = "\n";                       |
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
+            {1:~                                       }|*13
                                                     |
-          ]]}
+          ]],
+          }
         end,
       },
       {
@@ -1206,7 +1220,7 @@ int main()
             modifiers = {},
             type = 'comment',
             marked = true,
-          }
+          },
         },
         expected2 = {
           {
@@ -1288,10 +1302,11 @@ int main()
             modifiers = {},
             type = 'comment',
             marked = true,
-          }
+          },
         },
         expected_screen1 = function()
-          screen:expect{grid=[[
+          screen:expect {
+            grid = [[
             #include <iostream>                     |
                                                     |
             int {8:main}()                              |
@@ -1303,15 +1318,14 @@ int main()
             {6:    printf("%d\n", x);}                  |
             {6:#endif}                                  |
             ^}                                       |
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
+            {1:~                                       }|*4
                                                     |
-          ]]}
+          ]],
+          }
         end,
         expected_screen2 = function()
-          screen:expect{grid=[[
+          screen:expect {
+            grid = [[
             #include <iostream>                     |
                                                     |
             int {8:main}()                              |
@@ -1324,11 +1338,10 @@ int main()
             {6:    printf("%d\n", x);}                  |
             {6:^#endif}                                  |
             }                                       |
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
+            {1:~                                       }|*3
                                                     |
-          ]]}
+          ]],
+          }
         end,
       },
       {
@@ -1355,55 +1368,33 @@ int main()
             end_col = 6,
             type = 'variable',
             marked = true,
-          }
-        },
-        expected2 = {
+          },
         },
+        expected2 = {},
         expected_screen1 = function()
-          screen:expect{grid=[[
+          screen:expect {
+            grid = [[
             {7:string} = "test^"                         |
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
+            {1:~                                       }|*14
                                                     |
-          ]]}
+          ]],
+          }
         end,
         expected_screen2 = function()
-          screen:expect{grid=[[
+          screen:expect {
+            grid = [[
             ^                                        |
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
-            {1:~                                       }|
+            {1:~                                       }|*14
                                                     |
-          ]]}
+          ]],
+          }
         end,
       },
     }) do
       it(test.it, function()
         exec_lua(create_server_definition)
-        exec_lua([[
+        exec_lua(
+          [[
           local legend, resp1, resp2 = ...
           server = _create_server({
             capabilities = {
@@ -1431,7 +1422,11 @@ int main()
             semantic_tokens.stop(bufnr, client_id)
             semantic_tokens.start(bufnr, client_id, { debounce = 10 })
           end)
-        ]], test.legend, test.response1, test.response2)
+        ]],
+          test.legend,
+          test.response1,
+          test.response2
+        )
 
         insert(test.text1)
 
@@ -1446,11 +1441,14 @@ int main()
         if test.edit then
           feed(test.edit)
         else
-          exec_lua([[
+          exec_lua(
+            [[
             local text = ...
             vim.api.nvim_buf_set_lines(bufnr, 0, -1, false, vim.fn.split(text, "\n"))
             vim.wait(15) -- wait for debounce
-          ]], test.text2)
+          ]],
+            test.text2
+          )
         end
 
         test.expected_screen2()
