@@ -1,19 +1,21 @@
-local lfs = require('lfs')
-local helpers = require('test.unit.helpers')(after_each)
-local itp = helpers.gen_itp(it)
+local uv = vim.uv
+local t = require('test.unit.testutil')
+local itp = t.gen_itp(it)
 
-local cimport = helpers.cimport
-local eq = helpers.eq
-local neq = helpers.neq
-local ffi = helpers.ffi
-local cstr = helpers.cstr
-local to_cstr = helpers.to_cstr
-local NULL = helpers.NULL
-local OK = helpers.OK
-local FAIL = helpers.FAIL
+local cimport = t.cimport
+local eq = t.eq
+local neq = t.neq
+local ffi = t.ffi
+local cstr = t.cstr
+local to_cstr = t.to_cstr
+local NULL = t.NULL
+local OK = t.OK
+local FAIL = t.FAIL
+local mkdir = t.mkdir
 
 cimport('string.h')
 local cimp = cimport('./src/nvim/os/os.h', './src/nvim/path.h')
+local options = cimport('./src/nvim/option_vars.h')
 
 local length = 0
 local buffer = nil
@@ -21,11 +23,11 @@ local buffer = nil
 describe('path.c', function()
   describe('path_full_dir_name', function()
     setup(function()
-      lfs.mkdir('unit-test-directory')
+      mkdir('unit-test-directory')
     end)
 
     teardown(function()
-      lfs.rmdir('unit-test-directory')
+      uv.fs_rmdir('unit-test-directory')
     end)
 
     local function path_full_dir_name(directory, buf, len)
@@ -35,49 +37,55 @@ describe('path.c', function()
 
     before_each(function()
       -- Create empty string buffer which will contain the resulting path.
-      length = string.len(lfs.currentdir()) + 22
+      length = string.len(uv.cwd()) + 22
       buffer = cstr(length, '')
     end)
 
     itp('returns the absolute directory name of a given relative one', function()
       local result = path_full_dir_name('..', buffer, length)
       eq(OK, result)
-      local old_dir = lfs.currentdir()
-      lfs.chdir('..')
-      local expected = lfs.currentdir()
-      lfs.chdir(old_dir)
+      local old_dir = uv.cwd()
+      uv.chdir('..')
+      local expected = uv.cwd()
+      uv.chdir(old_dir)
       eq(expected, (ffi.string(buffer)))
     end)
 
     itp('returns the current directory name if the given string is empty', function()
       eq(OK, (path_full_dir_name('', buffer, length)))
-      eq(lfs.currentdir(), (ffi.string(buffer)))
-    end)
-
-    itp('fails if the given directory does not exist', function()
-      eq(FAIL, path_full_dir_name('does_not_exist', buffer, length))
+      eq(uv.cwd(), (ffi.string(buffer)))
     end)
 
     itp('works with a normal relative dir', function()
       local result = path_full_dir_name('unit-test-directory', buffer, length)
-      eq(lfs.currentdir() .. '/unit-test-directory', (ffi.string(buffer)))
+      eq(uv.cwd() .. '/unit-test-directory', (ffi.string(buffer)))
       eq(OK, result)
     end)
+
+    itp('works with a non-existing relative dir', function()
+      local result = path_full_dir_name('does-not-exist', buffer, length)
+      eq(uv.cwd() .. '/does-not-exist', (ffi.string(buffer)))
+      eq(OK, result)
+    end)
+
+    itp('fails with a non-existing absolute dir', function()
+      eq(FAIL, path_full_dir_name('/does_not_exist', buffer, length))
+    end)
   end)
 
   describe('path_full_compare', function()
-    local function path_full_compare(s1, s2, cn)
+    local function path_full_compare(s1, s2, cn, ee)
       s1 = to_cstr(s1)
       s2 = to_cstr(s2)
-      return cimp.path_full_compare(s1, s2, cn or 0)
+      return cimp.path_full_compare(s1, s2, cn or 0, ee or 1)
     end
 
     local f1 = 'f1.o'
     local f2 = 'f2.o'
     before_each(function()
       -- create the three files that will be used in this spec
-      io.open(f1, 'w').close()
-      io.open(f2, 'w').close()
+      io.open(f1, 'w'):close()
+      io.open(f2, 'w'):close()
     end)
 
     after_each(function()
@@ -176,7 +184,7 @@ describe('path.c', function()
 
     itp('returns the executable name of an invocation given a relative invocation', function()
       local invk, len = invocation_path_tail('directory/exe a b c')
-      compare("exe a b c", invk, len)
+      compare('exe a b c', invk, len)
       eq(3, len)
     end)
 
@@ -194,7 +202,7 @@ describe('path.c', function()
 
     itp('does not count arguments to the executable as part of its path', function()
       local invk, len = invocation_path_tail('exe a/b\\c')
-      compare("exe a/b\\c", invk, len)
+      compare('exe a/b\\c', invk, len)
       eq(3, len)
     end)
 
@@ -204,17 +212,17 @@ describe('path.c', function()
     end)
 
     itp('is equivalent to path_tail when args do not contain a path separator', function()
-      local ptail = cimp.path_tail(to_cstr("a/b/c x y z"))
+      local ptail = cimp.path_tail(to_cstr('a/b/c x y z'))
       neq(NULL, ptail)
       local tail = ffi.string(ptail)
-      local invk, _ = invocation_path_tail("a/b/c x y z")
+      local invk, _ = invocation_path_tail('a/b/c x y z')
       eq(tail, ffi.string(invk))
     end)
 
     itp('is not equivalent to path_tail when args contain a path separator', function()
-      local ptail = cimp.path_tail(to_cstr("a/b/c x y/z"))
+      local ptail = cimp.path_tail(to_cstr('a/b/c x y/z'))
       neq(NULL, ptail)
-      local invk, _ = invocation_path_tail("a/b/c x y/z")
+      local invk, _ = invocation_path_tail('a/b/c x y/z')
       neq((ffi.string(ptail)), (ffi.string(invk)))
     end)
   end)
@@ -262,27 +270,27 @@ describe('path.c', function()
 end)
 
 describe('path_try_shorten_fname', function()
-  local cwd = lfs.currentdir()
+  local cwd = uv.cwd()
 
   before_each(function()
-    lfs.mkdir('ut_directory')
+    mkdir('ut_directory')
   end)
 
   after_each(function()
-    lfs.chdir(cwd)
-    lfs.rmdir('ut_directory')
+    uv.chdir(cwd)
+    uv.fs_rmdir('ut_directory')
   end)
 
   describe('path_try_shorten_fname', function()
     itp('returns shortened path if possible', function()
-      lfs.chdir('ut_directory')
-      local full = to_cstr(lfs.currentdir() .. '/subdir/file.txt')
+      uv.chdir('ut_directory')
+      local full = to_cstr(uv.cwd() .. '/subdir/file.txt')
       eq('subdir/file.txt', (ffi.string(cimp.path_try_shorten_fname(full))))
     end)
 
     itp('returns `full_path` if a shorter version is not possible', function()
-      local old = lfs.currentdir()
-      lfs.chdir('ut_directory')
+      local old = uv.cwd()
+      uv.chdir('ut_directory')
       local full = old .. '/subdir/file.txt'
       eq(full, (ffi.string(cimp.path_try_shorten_fname(to_cstr(full)))))
     end)
@@ -294,14 +302,14 @@ describe('path_try_shorten_fname', function()
 end)
 
 describe('path.c path_guess_exepath', function()
-  local cwd = lfs.currentdir()
+  local cwd = uv.cwd()
 
-  for _,name in ipairs({'./nvim', '.nvim', 'foo/nvim'}) do
-    itp('"'..name..'" returns name catenated with CWD', function()
+  for _, name in ipairs({ './nvim', '.nvim', 'foo/nvim' }) do
+    itp('"' .. name .. '" returns name catenated with CWD', function()
       local bufsize = 255
       local buf = cstr(bufsize, '')
       cimp.path_guess_exepath(name, buf, bufsize)
-      eq(cwd..'/'..name, ffi.string(buf))
+      eq(cwd .. '/' .. name, ffi.string(buf))
     end)
   end
 
@@ -323,10 +331,10 @@ describe('path.c path_guess_exepath', function()
 
   itp('does not crash if $PATH item exceeds MAXPATHL', function()
     local orig_path_env = os.getenv('PATH')
-    local name = 'cat'  -- Some executable in $PATH.
+    local name = 'cat' -- Some executable in $PATH.
     local bufsize = 255
     local buf = cstr(bufsize, '')
-    local insane_path = orig_path_env..':'..(("x/"):rep(4097))
+    local insane_path = orig_path_env .. ':' .. (('x/'):rep(4097))
 
     cimp.os_setenv('PATH', insane_path, true)
     cimp.path_guess_exepath(name, buf, bufsize)
@@ -337,7 +345,7 @@ describe('path.c path_guess_exepath', function()
   end)
 
   itp('returns full path found in $PATH', function()
-    local name = 'cat'  -- Some executable in $PATH.
+    local name = 'cat' -- Some executable in $PATH.
     local bufsize = 255
     local buf = cstr(bufsize, '')
     cimp.path_guess_exepath(name, buf, bufsize)
@@ -348,8 +356,8 @@ end)
 
 describe('path.c', function()
   setup(function()
-    lfs.mkdir('unit-test-directory');
-    io.open('unit-test-directory/test.file', 'w').close()
+    mkdir('unit-test-directory')
+    io.open('unit-test-directory/test.file', 'w'):close()
 
     -- Since the tests are executed, they are called by an executable. We use
     -- that executable for several asserts.
@@ -357,12 +365,12 @@ describe('path.c', function()
 
     -- Split absolute_executable into a directory and the actual file name for
     -- later usage.
-    local directory, executable_name = string.match(absolute_executable, '^(.*)/(.*)$')  -- luacheck: ignore
+    local directory, executable_name = string.match(absolute_executable, '^(.*)/(.*)$') -- luacheck: ignore
   end)
 
   teardown(function()
     os.remove('unit-test-directory/test.file')
-    lfs.rmdir('unit-test-directory')
+    uv.fs_rmdir('unit-test-directory')
   end)
 
   describe('vim_FullName', function()
@@ -372,8 +380,8 @@ describe('path.c', function()
       return buf, result
     end
 
-    local function get_buf_len(s, t)
-      return math.max(string.len(s), string.len(t)) + 1
+    local function get_buf_len(s, q)
+      return math.max(string.len(s), string.len(q)) + 1
     end
 
     itp('fails if given filename is NULL', function()
@@ -405,16 +413,19 @@ describe('path.c', function()
     end)
 
     itp('fails and uses filename if given filename contains non-existing directory', function()
-      local filename = 'non_existing_dir/test.file'
-      local buflen = string.len(filename) + 1
-      local do_expand = 1
-      local buf, result = vim_FullName(filename, buflen, do_expand)
-      eq(filename, ffi.string(buf))
-      eq(FAIL, result)
+      -- test with different filename lengths
+      for rep = 1, 10 do
+        local filename = ('non_existing_'):rep(rep) .. 'dir/test.file'
+        local buflen = string.len(filename) + 1
+        local do_expand = 1
+        local buf, result = vim_FullName(filename, buflen, do_expand)
+        eq(filename, ffi.string(buf))
+        eq(FAIL, result)
+      end
     end)
 
     itp('concatenates filename if it does not contain a slash', function()
-      local expected = lfs.currentdir() .. '/test.file'
+      local expected = uv.cwd() .. '/test.file'
       local filename = 'test.file'
       local buflen = get_buf_len(expected, filename)
       local do_expand = 1
@@ -424,7 +435,7 @@ describe('path.c', function()
     end)
 
     itp('concatenates directory name if it does not contain a slash', function()
-      local expected = lfs.currentdir() .. '/..'
+      local expected = uv.cwd() .. '/..'
       local filename = '..'
       local buflen = get_buf_len(expected, filename)
       local do_expand = 1
@@ -433,18 +444,21 @@ describe('path.c', function()
       eq(OK, result)
     end)
 
-    itp('enters given directory (instead of just concatenating the strings) if possible and if path contains a slash', function()
-      local old_dir = lfs.currentdir()
-      lfs.chdir('..')
-      local expected = lfs.currentdir() .. '/test.file'
-      lfs.chdir(old_dir)
-      local filename = '../test.file'
-      local buflen = get_buf_len(expected, filename)
-      local do_expand = 1
-      local buf, result = vim_FullName(filename, buflen, do_expand)
-      eq(expected, ffi.string(buf))
-      eq(OK, result)
-    end)
+    itp(
+      'enters given directory (instead of just concatenating the strings) if possible and if path contains a slash',
+      function()
+        local old_dir = uv.cwd()
+        uv.chdir('..')
+        local expected = uv.cwd() .. '/test.file'
+        uv.chdir(old_dir)
+        local filename = '../test.file'
+        local buflen = get_buf_len(expected, filename)
+        local do_expand = 1
+        local buf, result = vim_FullName(filename, buflen, do_expand)
+        eq(expected, ffi.string(buf))
+        eq(OK, result)
+      end
+    )
 
     itp('just copies the path if it is already absolute and force=0', function()
       local absolute_path = '/absolute/path'
@@ -456,7 +470,7 @@ describe('path.c', function()
     end)
 
     itp('fails and uses filename when the path is relative to HOME', function()
-      eq(false, cimp.os_isdir('~')) -- sanity check: no literal "~" directory.
+      eq(false, cimp.os_isdir('~'), 'sanity check: no literal "~" directory')
       local absolute_path = '~/home.file'
       local buflen = string.len(absolute_path) + 1
       local do_expand = 1
@@ -466,7 +480,7 @@ describe('path.c', function()
     end)
 
     itp('works with some "normal" relative path with directories', function()
-      local expected = lfs.currentdir() .. '/unit-test-directory/test.file'
+      local expected = uv.cwd() .. '/unit-test-directory/test.file'
       local filename = 'unit-test-directory/test.file'
       local buflen = get_buf_len(expected, filename)
       local do_expand = 1
@@ -476,7 +490,7 @@ describe('path.c', function()
     end)
 
     itp('does not modify the given filename', function()
-      local expected = lfs.currentdir() .. '/unit-test-directory/test.file'
+      local expected = uv.cwd() .. '/unit-test-directory/test.file'
       local filename = to_cstr('unit-test-directory/test.file')
       local buflen = string.len(expected) + 1
       local buf = cstr(buflen, '')
@@ -498,8 +512,18 @@ describe('path.c', function()
       eq(OK, result)
     end)
 
+    itp('does not remove trailing slash from non-existing relative directory #20847', function()
+      local expected = uv.cwd() .. '/non_existing_dir/'
+      local filename = 'non_existing_dir/'
+      local buflen = get_buf_len(expected, filename)
+      local do_expand = 1
+      local buf, result = vim_FullName(filename, buflen, do_expand)
+      eq(expected, ffi.string(buf))
+      eq(OK, result)
+    end)
+
     itp('expands "./" to the current directory #7117', function()
-      local expected = lfs.currentdir() .. '/unit-test-directory/test.file'
+      local expected = uv.cwd() .. '/unit-test-directory/test.file'
       local filename = './unit-test-directory/test.file'
       local buflen = get_buf_len(expected, filename)
       local do_expand = 1
@@ -509,7 +533,7 @@ describe('path.c', function()
     end)
 
     itp('collapses "foo/../foo" to "foo" #7117', function()
-      local expected = lfs.currentdir() .. '/unit-test-directory/test.file'
+      local expected = uv.cwd() .. '/unit-test-directory/test.file'
       local filename = 'unit-test-directory/../unit-test-directory/test.file'
       local buflen = get_buf_len(expected, filename)
       local do_expand = 1
@@ -526,8 +550,12 @@ describe('path.c', function()
       return ffi.string(c_file)
     end
 
-    before_each(function() lfs.mkdir('CamelCase') end)
-    after_each(function() lfs.rmdir('CamelCase') end)
+    before_each(function()
+      mkdir('CamelCase')
+    end)
+    after_each(function()
+      uv.fs_rmdir('CamelCase')
+    end)
 
     if ffi.os == 'Windows' or ffi.os == 'OSX' then
       itp('Corrects the case of file names in Mac and Windows', function()
@@ -547,14 +575,14 @@ describe('path.c', function()
       local path1 = cstr(100, 'path1')
       local to_append = to_cstr('path2')
       eq(OK, (cimp.append_path(path1, to_append, 100)))
-      eq("path1/path2", (ffi.string(path1)))
+      eq('path1/path2', (ffi.string(path1)))
     end)
 
     itp('joins given paths without adding an unnecessary slash', function()
       local path1 = cstr(100, 'path1/')
       local to_append = to_cstr('path2')
       eq(OK, cimp.append_path(path1, to_append, 100))
-      eq("path1/path2", (ffi.string(path1)))
+      eq('path1/path2', (ffi.string(path1)))
     end)
 
     itp('fails and uses filename if there is not enough space left for to_append', function()
@@ -592,15 +620,61 @@ describe('path.c', function()
     end
 
     itp('returns true if filename starts with a slash', function()
-      eq(OK, path_is_absolute('/some/directory/'))
+      eq(true, path_is_absolute('/some/directory/'))
     end)
 
     itp('returns true if filename starts with a tilde', function()
-      eq(OK, path_is_absolute('~/in/my/home~/directory'))
+      eq(true, path_is_absolute('~/in/my/home~/directory'))
     end)
 
     itp('returns false if filename starts not with slash nor tilde', function()
-      eq(FAIL, path_is_absolute('not/in/my/home~/directory'))
+      eq(false, path_is_absolute('not/in/my/home~/directory'))
+    end)
+  end)
+
+  describe('path_with_extension', function()
+    local function path_with_extension(filename, extension)
+      local c_filename = to_cstr(filename)
+      local c_extension = to_cstr(extension)
+      return cimp.path_with_extension(c_filename, c_extension)
+    end
+
+    itp('returns true if filename includes a provided extension', function()
+      eq(true, path_with_extension('/some/path/file.lua', 'lua'))
+    end)
+
+    itp('returns false if filename does not include a provided extension', function()
+      eq(false, path_with_extension('/some/path/file.vim', 'lua'))
+      eq(false, path_with_extension('/some/path/file', 'lua'))
+    end)
+
+    itp("respects 'fileignorecase' option", function()
+      options.p_fic = false
+      eq(false, path_with_extension('/some/path/file.VIM', 'vim'))
+      eq(false, path_with_extension('/some/path/file.LUA', 'lua'))
+      options.p_fic = true
+      eq(true, path_with_extension('/some/path/file.VIM', 'vim'))
+      eq(true, path_with_extension('/some/path/file.LUA', 'lua'))
+    end)
+  end)
+
+  describe('path_with_url', function()
+    itp('scheme is alpha and inner hyphen only', function()
+      local function path_with_url(fname)
+        return cimp.path_with_url(to_cstr(fname))
+      end
+      eq(1, path_with_url([[test://xyz/foo/b0]]))
+      eq(2, path_with_url([[test:\\xyz\foo\b0]]))
+      eq(0, path_with_url([[test+abc://xyz/foo/b1]]))
+      eq(0, path_with_url([[test_abc://xyz/foo/b2]]))
+      eq(1, path_with_url([[test-abc://xyz/foo/b3]]))
+      eq(2, path_with_url([[test-abc:\\xyz\foo\b3]]))
+      eq(0, path_with_url([[-test://xyz/foo/b4]]))
+      eq(0, path_with_url([[test-://xyz/foo/b5]]))
+      eq(1, path_with_url([[test-C:/xyz/foo/b5]]))
+      eq(1, path_with_url([[test-custom:/xyz/foo/b5]]))
+      eq(0, path_with_url([[c:/xyz/foo/b5]]))
+      eq(0, path_with_url([[C:/xyz/foo/b5]]))
     end)
   end)
 end)
